#ifndef LINT
static char *rcsid="$Id: contentfilter.c,v 1.12 2002/09/05 13:36:21 crosser Exp $";
#endif

/*
	$Log: contentfilter.c,v $
	Revision 1.12  2002/09/05 13:36:21  crosser
	spell checking, more commentary, for release 1.00
	
	Revision 1.11  2002/08/26 16:02:59  crosser
	made different patterns for envelope, headers and body
	
	Revision 1.10  2002/08/26 12:15:24  crosser
	Oops, was reading from wrong file descriptor
	
	Revision 1.9  2002/08/26 10:31:44  crosser
	got rid of stdio in contentfilter
	
	Revision 1.8  2002/08/25 19:05:34  crosser
	ready version 0.01
	
	Revision 1.7  2002/08/25 14:02:14  crosser
	implement special characters
	include README
	make test script
	
	Revision 1.6  2002/08/25 11:22:51  crosser
	work in progress
	
	Revision 1.5  2002/08/23 22:00:50  crosser
	Multiple patterns
	
	Revision 1.4  2002/08/23 16:26:06  crosser
	working prototype
	
	Revision 1.3  2002/04/25 13:51:45  crosser
	try mmap
	
	Revision 1.2  2002/04/25 12:56:03  crosser
	fix comment
	
	Revision 1.1.1.1  2002/04/25 12:53:35  crosser
	Contentfilter sources

*/

/*
	Copyright (c) 2002 Eugene Crosser

	You may distribute and/or use for any purpose modified or unmodified
	copies of this software if you preserve the copyright notice above.

	THIS SOFTWARE IS PROVIDED AS IS AND COME WITH NO WARRANTY OF ANY
	KIND, EITHER EXPRESSED OR IMPLIED.  IN NO EVENT WILL THE
	COPYRIGHT HOLDER BE LIABLE FOR ANY DAMAGES RESULTING FROM THE
	USE OF THIS SOFTWARE.

	Lean and mean smtp-contentfilter for Zmailer.  Designed to be
	low resource consuming.  Configurability, flexibility and even
	portability are sacrificed in sake of speed and size.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#ifdef DEBUG
# include <stdio.h>
#endif

#define NL '\n'
#define CR '\r'

static char *rulefile=RULEFILE;

static char *ok=	"0 250 2.6.0 Accepted\n";
static char *begone=	"-1 550 5.7.1 Policy rejection\n";

/*
	Simplified version of perror/fprintf/syslog.  Gets format
	string that can contain text, '%%', '%s' and '%m'.  Outputs
	to stderror bypassing stdio.
*/

#define ERRFD 2

static void
myerror(char *fmt, ...)
{
	va_list ap;
	int pcseen=0;
	int count=0;
	char buf[1024];
	char *p=buf,*s=NULL;

	va_start(ap, fmt);
	while (*fmt) {
		if (count >= (sizeof(buf)-1)) break;
		if (pcseen) {
			pcseen=0;
			if ((*fmt == 's') || (*fmt == 'm')) {
				switch (*fmt) {
				case 's':
					s=va_arg(ap,char*);
					break;
				case 'm':
					s=strerror(errno);
					break; 
				}
				while (*s) {
					*p++=*s++;
					count++;
					if (count >= (sizeof(buf)-1)) break;
				}
			} else {
				*p++=*fmt;
				count++;
			}
		} else {
			if (*fmt == '%') pcseen=1;
			else {
				*p++=*fmt;
				count++;
			}
		}
		fmt++;
	}
	va_end(ap);
	*p='\0';
	(void)write(ERRFD,buf,count);
}

typedef struct _pats {
	int patc;
	char **patv;
} pats_t;

/*
	scanmsg().  The real job of scanning is done here.
	Args: start and size of memory to scan, array of
	patterns to match passed in three groups a la argc, argv.
	Ret: message string.
*/

static char *
scanmsg(char *msgdata, size_t msgsize, pats_t *pats)
{
	int i,j;
	char *p;
	char **anc;
	char **ptr;
	int maxcnt=0;

	for (j=0;j<3;j++) if (maxcnt < pats[j].patc) maxcnt=pats[j].patc;
	if (((anc=(char**)alloca(maxcnt*sizeof(char*))) == NULL) ||
	    ((ptr=(char**)alloca(maxcnt*sizeof(char*))) == NULL)) {
		myerror("alloca: %m\n");
		return "0 Oops.. filter cannot alloca arrays\n";
	}

	p=msgdata;
	for (j=0;j<3;j++) {
		int patc=pats[j].patc;
		char **patv=pats[j].patv;

#ifdef DEBUG
		fprintf(stderr,"section %d starts at offset %d\n",
				j,p-msgdata);
#endif

		for (i=0;i<patc;i++) anc[i]=ptr[i]=patv[i];
	
		for (;p<msgdata+msgsize;p++) {
			if (*p == CR) continue;
			for (i=0;i<patc;i++) {
				if (*(ptr[i]) == '\0') {
#ifdef DEBUG
		fprintf(stderr,"match \"%s\" (%d) in section %d at offset %d\n",
				patv[i],i,j,p-msgdata);
#endif
					/* pattern zero has the meaning
					   "end of section" for sections
					   "envelope" and "headers" */
					if ((i == 0) && (j < 2)) goto nextsec;
					else return begone;
				}
				if (*p == NL) anc[i]=patv[i];
				if (*p == *(ptr[i])) {
					ptr[i]++;
					if (*(ptr[i]) == CR) {
						ptr[i]++;
						anc[i]=ptr[i];
					}
				}
				else ptr[i]=anc[i];
			}
		}
		nextsec:
	}
	return ok;
}

/*
	process(). Open and mmap message file and call scanmsg() upon it
	Args: char * with filename, pattern arrays.
	Ret: message string.
*/

static inline char *
process(char *msgfile, pats_t *pats)
{
	int msgfd;
	char *msgdata;
	struct stat stbuf;
	off_t msgsize;
	char *rc;

	if ((msgfd=open(msgfile,O_RDONLY)) == -1) {
		myerror("open %s: %m\n",msgfile);
		return "0 Oops.. filter can't open message file\n";
	}
	if (fstat(msgfd,&stbuf) == -1) {
		myerror("fstat %s: %m\n",msgfile);
		return "0 Oops.. filter can't fstat message file\n";
	}
	msgsize=stbuf.st_size;
	if (msgsize > MAXSIZE) msgsize=MAXSIZE;
	if ((msgdata=(char*)mmap((void*)0,msgsize,PROT_READ,
				MAP_SHARED,msgfd,(off_t)0)) == NULL) {
		myerror("mmap %s: %m\n",msgfile);
		return "0 Oops.. filter can't mmap message file\n";
	}

	rc=scanmsg(msgdata,msgsize,pats);

	if (munmap(msgdata,msgsize) == -1) {
		myerror("munmap %s: %m\n",msgfile);
		return "0 Oops.. filter can't munmap message file\n";
	}
	if ((close(msgfd)) == -1) {
		myerror("close %s: %m\n",msgfile);
		return "0 Oops.. filter can't close message file\n";
	}
	return rc;
}

/*
	main().  Master loop.  Open, mmap and relocate patterns file
	and start looping on '#hungry' - process - response.
*/

int
main(int argc, char *argv[])
{
	int rulefd;
	char *ruledata;
	int *rulecount;
	size_t *ruleoffset;
	struct stat stbuf;
	off_t rulesize;
	char msgfile[128];
	ssize_t readsize;
	char *p,*rc;
	int i,j;
	pats_t pats[3];

	if ((rulefd=open(rulefile,O_RDONLY)) == -1) {
		myerror("open %s: %m\n",rulefile);
		exit(0);
	}
	if (fstat(rulefd,&stbuf) == -1) {
		myerror("fstat %s: %m\n",rulefile);
		exit(0);
	}
	rulesize=stbuf.st_size;
	if ((ruledata=(char*)mmap((void*)0,rulesize,PROT_READ,
				MAP_PRIVATE,rulefd,(off_t)0)) == NULL) {
		myerror("mmap %s: %m\n",rulefile);
		exit(0);
	}

	rulecount=(int*)ruledata;
	ruleoffset=(size_t*)(rulecount+3);
	for (j=0;j<3;j++) {
		pats[j].patc=rulecount[j];
		pats[j].patv=(char**)alloca(pats[j].patc*sizeof(char*));
		if (pats[j].patv == NULL) {
			myerror("alloca space for patv: %m\n");
			exit(0);
		}
		for (i=0;i<pats[j].patc;i++)
			pats[j].patv[i]=
				ruledata+*ruleoffset++;
	}
#ifdef DEBUG
	for (j=0;j<3;j++) {
		fprintf(stderr,"section %d (%d entries):\n",j,pats[j].patc);
		for (i=0;i<pats[j].patc;i++) {
			fprintf(stderr,"  patv[%d]=\"%s\"\n",
						i,pats[j].patv[i]);
		}
	}
#endif

	while (1) {
		if (write(1,"#hungry\n",8) != 8) {
			myerror("writing #hungry: %m\n");
			exit(1);
		}
		if ((readsize=read(0,msgfile,sizeof(msgfile)-1)) <= 0) {
			break;
		}
		msgfile[readsize]='\0';
		for (p=msgfile+readsize-1;
		     (p > msgfile) && ((*p == '\r') || (*p == '\n'));
		     p--) *p='\0';
		rc=process(msgfile,pats);
		if (write(1,rc,strlen(rc)) != strlen(rc)) {
			myerror("writing response string: %m\n");
			exit(1);
		}
	}

	if (munmap(ruledata,rulesize) == -1) {
		myerror("munmap %s: %m\n",rulefile);
		exit(0);
	}
	if ((close(rulefd)) == -1) {
		myerror("close %s: %m\n",rulefile);
		exit(0);
	}

	return 0;
}
